/*
* ltokens.c
* lexer library -- uses private functions (politely)
* Luiz Henrique de Figueiredo <lhf@tecgraf.puc-rio.br>
* 22 Jan 2004 13:32:50
* This code is hereby placed in the public domain.
*/

#include <errno.h>
#include <string.h>
#include <stdio.h>

#include "lua.h"
#include "lauxlib.h"

#include "llex.h"
#include "lmem.h"
#include "lstring.h"
#include "lzio.h"

static int dolex(lua_State *L, lua_Chunkreader reader, void *data, const char *name)
{
 int n=0;
 ZIO z;
 Mbuffer B;
 LexState X;
 luaZ_init(&z,reader,data,name);
 luaZ_initbuffer(L,&B);
 X.buff=&B;
 luaX_setinput(L,&X,&z,luaS_new(L,name));
 lua_newtable(L);
 for (;;)
 {
  int t=luaX_lex(&X,&X.t.seminfo);
  if (t==TK_EOS)
  {
#ifdef ADDN
   lua_pushliteral(L,"n");
   lua_pushnumber(L,n);
   lua_settable(L,-3);
#endif
   luaZ_freebuffer(L,&B);
   return 1;
  }
  lua_pushnumber(L,++n);
  lua_newtable(L);

  lua_pushliteral(L,"line");
  lua_pushnumber(L,X.linenumber);
  lua_settable(L,-3);

#ifdef ADDRAW
  lua_pushliteral(L,"TOKEN");
  lua_pushnumber(L,t);
  lua_settable(L,-3);

  lua_pushliteral(L,"TEXT");
  lua_pushstring(L,X.buff->buffer);
  lua_settable(L,-3);
#endif

  lua_pushliteral(L,"token");
  switch (t)
  {
   case TK_STRING:
    lua_pushliteral(L,"string");
    lua_settable(L,-3);
    lua_pushliteral(L,"value");
    lua_pushstring(L,getstr(X.t.seminfo.ts));
    lua_settable(L,-3);
    break;
   case TK_NAME:
    lua_pushliteral(L,"name");
    lua_settable(L,-3);
    lua_pushliteral(L,"value");
    lua_pushstring(L,getstr(X.t.seminfo.ts));
    lua_settable(L,-3);
    break;
   case TK_NUMBER:
    lua_pushliteral(L,"number");
    lua_settable(L,-3);
    lua_pushliteral(L,"value");
    lua_pushnumber(L,X.t.seminfo.r);
    lua_settable(L,-3);
    lua_pushliteral(L,"text");
    lua_pushstring(L,X.buff->buffer);
    lua_settable(L,-3);
    break;
   default:
   {
    const char *s=luaX_token2str(&X,t);
    if (t<FIRST_RESERVED) ; else lua_pushstring(L,s);
    lua_settable(L,-3);
    break;
   }
  }
  lua_settable(L,-3);
 }
}

static const char *freader(lua_State *L, void *data, size_t *size)
{
 static char b[512];
 FILE *f=(FILE*) data;
 if (feof(f)) return NULL;
 *size=fread(b,1,sizeof(b),f);
 return (*size > 0) ? b : NULL;
}

static int ftokens(lua_State *L)
{
 FILE *f;
 const char *filename;
 int rc;
 if (lua_isnone(L,1))
 {
  filename="=stdin";
  f=stdin;
 }
 else
 {
  filename=luaL_checkstring(L,1);
  f=fopen(filename,"r");
  if (f==NULL)
  {
   lua_pushnil(L);
   lua_pushfstring(L,"%s: %s",filename,strerror(errno));
   lua_pushnumber(L,errno);
   return 3;
  }
  filename=lua_pushfstring(L,"@%s",filename);
 }
 rc=dolex(L,freader,f,filename);
 if (f!=stdin) fclose(f);
 return rc;
}

typedef struct LoadS {
  const char *s;
  size_t size;
} LoadS;

static const char *sreader(lua_State *L, void *data, size_t *size)
{
 LoadS *S=(LoadS*)data;
 const char *t=S->s;
 (void)L;
 *size=S->size;
 S->s=NULL;
 S->size=0;
 return t;
}

static int stokens(lua_State *L)
{
 LoadS S;
 S.s=luaL_checklstring(L,1,&S.size);
 return dolex(L,sreader,&S,S.s);
}

static int ptokens(lua_State *L)
{
 lua_pushvalue(L,lua_upvalueindex(1));
 lua_insert(L,1);
 if (lua_pcall(L,1,LUA_MULTRET,0))
 {
  lua_pushnil(L);
  lua_insert(L,-2);
  return 2;
 }
 return lua_gettop(L);
}

LUALIB_API int luaopen_tokens(lua_State *L)
{
 lua_pushcclosure(L,ftokens,0);
 lua_pushcclosure(L,ptokens,1);
 lua_setglobal(L,"ftokens");
 lua_pushcclosure(L,stokens,0);
 lua_pushcclosure(L,ptokens,1);
 lua_setglobal(L,"stokens");
 return 0;
}
